from enum import IntEnum


class SHGeoType(IntEnum):
    unknown = 0
    zone = 1
    cyl = 2
    msh = 3
    plane = 4
    dzone = 5
    dcyl = 6
    dmsh = 7
    dplane = 8
    dcylz = 10
    dmshz = 11
    trace = 13
    voxscore = 14
    geomap = 15

    def __str__(self):
        return self.name.upper()



import logging

import numpy as np

from pymchelper.axis import MeshAxis
from pymchelper.page import Page
from pymchelper.readers.shieldhit.reader_base import SHReader, mesh_unit_and_name, _bintyp, _get_detector_unit, \
    read_next_token
from pymchelper.readers.shieldhit.binary_spec import SHBDOTagID, detector_name_from_bdotag
from pymchelper.shieldhit.detector.estimator_type import SHGeoType
from pymchelper.shieldhit.detector.detector_type import SHDetType

logger = logging.getLogger(__name__)


class SHReaderBDO2016(SHReader):
    """
    Binary format reader from version >= 0.6
    This format doesn't have support for multiple pages per estimator
    """
    def read_data(self, estimator):
        logger.debug("Reading: " + self.filename)
        with open(self.filename, "rb") as f:
            d1 = np.dtype([('magic', 'S6'),
                           ('end', 'S2'),
                           ('vstr', 'S16')])

            _x = np.fromfile(f, dtype=d1, count=1)  # read the data into numpy
            logger.debug("Magic : " + _x['magic'][0].decode('ASCII'))
            logger.debug("Endian: " + _x['end'][0].decode('ASCII'))
            logger.debug("VerStr: " + _x['vstr'][0].decode('ASCII'))

            # if no pages are present, add first one
            if not estimator.pages:
                estimator.add_page(Page())

            while f:
                token = read_next_token(f)
                if token is None:
                    break

                pl_id, _pl_type, _pl_len, _pl = token

                pl = [None] * _pl_len

                # decode all strings (currently there will never be more than one per token)
                if 'S' in _pl_type.decode('ASCII'):
                    for i, _j in enumerate(_pl):
                        pl[i] = _pl[i].decode('ASCII').strip()
                else:
                    pl = _pl

                try:
                    token_name = SHBDOTagID(pl_id).name
                    logger.debug("Read token {:s} (0x{:02x}) value {} type {:s} length {:d}".format(
                        token_name,
                        pl_id,
                        _pl,
                        _pl_type.decode('ASCII'),
                        _pl_len
                    ))
                except ValueError:
                    logger.info("Skipping token (0x{:02x}) value {} type {:s} length {:d}".format(
                        pl_id,
                        _pl,
                        _pl_type.decode('ASCII'),
                        _pl_len
                    ))

                if SHBDOTagID.shversion == pl_id:
                    estimator.mc_code_version = pl[0]
                    logger.debug("MC code version:" + estimator.mc_code_version)

                if SHBDOTagID.filedate == pl_id:
                    estimator.filedate = pl[0]

                if SHBDOTagID.user == pl_id:
                    estimator.user = pl[0]

                if SHBDOTagID.host == pl_id:
                    estimator.host = pl[0]

                if SHBDOTagID.rt_nstat == pl_id:
                    estimator.number_of_primaries = pl[0]

                # beam configuration etc...
                if pl_id in detector_name_from_bdotag:
                    setattr(estimator, detector_name_from_bdotag[pl_id], pl[0])

                # estimator block here ---
                if SHBDOTagID.det_geotyp == pl_id:
                    estimator.geotyp = SHGeoType[pl[0].strip().lower()]

                if SHBDOTagID.ext_ptvdose == pl_id:
                    estimator.tripdose = 0.0

                if SHBDOTagID.ext_nproj == pl_id:
                    estimator.tripntot = -1

                # read a single detector
                if SHBDOTagID.det_dtype == pl_id:
                    estimator.pages[0].dettyp = SHDetType(pl[0])

                if SHBDOTagID.det_part == pl_id:  # particle to be scored
                    estimator.scored_particle_code = pl[0]
                if SHBDOTagID.det_partz == pl_id:  # particle to be scored
                    estimator.scored_particle_z = pl[0]
                if SHBDOTagID.det_parta == pl_id:  # particle to be scored
                    estimator.scored_particle_a = pl[0]

                if SHBDOTagID.det_nbin == pl_id:
                    nx = pl[0]
                    ny = pl[1]
                    nz = pl[2]

                if SHBDOTagID.det_xyz_start == pl_id:
                    xmin = pl[0]
                    ymin = pl[1]
                    zmin = pl[2]

                if SHBDOTagID.det_xyz_stop == pl_id:
                    xmax = pl[0]
                    ymax = pl[1]
                    zmax = pl[2]

                # partial support for differential scoring (only linear binning)
                # TODO add some support for DMSH, DCYL and DZONE
                # TODO add support for logarithmic binning
                diff_geotypes = {SHGeoType.dplane, SHGeoType.dmsh, SHGeoType.dcyl, SHGeoType.dzone}
                if hasattr(estimator, 'geotyp') and estimator.geotyp in diff_geotypes:
                    if SHBDOTagID.det_dif_start == pl_id:
                        estimator.dif_min = pl[0]

                    if SHBDOTagID.det_dif_stop == pl_id:
                        estimator.dif_max = pl[0]

                    if SHBDOTagID.det_nbine == pl_id:
                        estimator.dif_n = pl[0]

                    if SHBDOTagID.det_difftype == pl_id:
                        estimator.dif_type = pl[0]

                if SHBDOTagID.det_zonestart == pl_id:
                    estimator.zone_start = pl[0]

                if SHBDOTagID.data_block == pl_id:
                    estimator.pages[0].data_raw = np.asarray(pl)

            # TODO: would be better to not overwrite x,y,z and make proper case for ZONE scoring later.
            if estimator.geotyp in {SHGeoType.zone, SHGeoType.dzone}:
                # special case for zone scoring, x min and max will be zone numbers
                xmin = estimator.zone_start
                xmax = xmin + nx - 1
                ymin = 0.0
                ymax = 0.0
                zmin = 0.0
                zmax = 0.0
            elif estimator.geotyp in {SHGeoType.plane, SHGeoType.dplane}:
                # special case for plane scoring, according to documentation we have:
                #  xmin, ymin, zmin = Sx, Sy, Sz (point on the plane)
                #  xmax, ymax, zmax = nx, ny, nz (normal vector)
                # to avoid situation where i.e. xmax < xmin (corresponds to nx < Sx)
                # we store only point on the plane
                estimator.sx, estimator.sy, estimator.sz = xmin, ymin, zmin
                estimator.nx, estimator.ny, estimator.nz = xmax, ymax, zmax
                xmax = xmin
                ymax = ymin
                zmax = zmin

            # check if scoring quantity is LET, if yes, than change units from [MeV/cm] to [keV/um]
            if hasattr(estimator, 'dif_type') and estimator.dif_type == 2:
                estimator.dif_min /= 10.0
                estimator.dif_max /= 10.0

            # # differential scoring data replacement
            if hasattr(estimator, 'dif_min') and hasattr(estimator, 'dif_max') and hasattr(estimator, 'dif_n'):
                if nz == 1:
                    # max two axis (X or Y) filled with scored value, Z axis empty
                    # we can put differential quantity as Z axis
                    nz = estimator.dif_n
                    zmin = estimator.dif_min
                    zmax = estimator.dif_max
                    estimator.dif_axis = 2
                elif ny == 1:
                    # Z axis filled with scored value (X axis maybe also), Y axis empty
                    # we can put differential quantity as Y axis
                    ny = estimator.dif_n
                    ymin = estimator.dif_min
                    ymax = estimator.dif_max
                    estimator.dif_axis = 1
                elif nx == 1:
                    nx = estimator.dif_n
                    xmin = estimator.dif_min
                    xmax = estimator.dif_max
                    estimator.dif_axis = 0

            xunit, xname = mesh_unit_and_name(estimator, 0)
            yunit, yname = mesh_unit_and_name(estimator, 1)
            zunit, zname = mesh_unit_and_name(estimator, 2)

            estimator.x = MeshAxis(n=np.abs(nx),
                                   min_val=xmin,
                                   max_val=xmax,
                                   name=xname,
                                   unit=xunit,
                                   binning=_bintyp(nx))
            estimator.y = MeshAxis(n=np.abs(ny),
                                   min_val=ymin,
                                   max_val=ymax,
                                   name=yname,
                                   unit=yunit,
                                   binning=_bintyp(ny))
            estimator.z = MeshAxis(n=np.abs(nz),
                                   min_val=zmin,
                                   max_val=zmax,
                                   name=zname,
                                   unit=zunit,
                                   binning=_bintyp(nz))

            estimator.pages[0].unit, estimator.pages[0].name = _get_detector_unit(estimator.pages[0].dettyp,
                                                                                  estimator.geotyp)

            estimator.file_format = 'bdo2016'

            logger.debug("Done reading bdo file.")
            logger.debug("Detector data : " + str(estimator.pages[0].data))
            logger.debug("Detector nstat: " + str(estimator.number_of_primaries))
            logger.debug("Detector nx   : " + str(estimator.x.n))
            logger.debug("Detector ny   : " + str(estimator.y.n))
            logger.debug("Detector nz   : " + str(estimator.z.n))
            estimator.file_counter = 1
        super(SHReaderBDO2016, self).read_data(estimator)
        return True



import logging

import numpy as np

from pymchelper.axis import MeshAxis
from pymchelper.page import Page
from pymchelper.readers.common import ReaderFactory, Reader
from pymchelper.flair.Data import Usrbin, UsrTrack, unpackArray, Usrbdx, Resnuclei, Usrxxx

logger = logging.getLogger(__name__)


class FlukaReaderFactory(ReaderFactory):
    """
    Class responsible for discovery of filetype.
    """
    def get_reader(self):
        """
        Try reading header of Fluka binary file and return a corresponding FlukaReader object

        :return: FlukaReader class if file is digested by Usrxxx Flair reader. None is returned otherwise
        """
        try:
            Usrxxx(self.filename)
            return FlukaReader
        except IOError:
            pass
        return None


class FlukaReader(Reader):

    @property
    def corename(self):
        """
        Fluka output filenames follow this pattern: corename_fort.XX.
        :return: corename part of output file or None in case filename doesn't follow Fluka naming pattern
        """
        core_name = None
        if "_fort" in self.filename:
            core_name = self.filename[-2:]
        return core_name

    def parse_usrbin(self, estimator):
        """
        USRBIN scores distribution of one of several quantities in a regular spatial
        structure (binning detector) independent from the geometry.
        :param estimator: an Estimator object, will be modified here and filled with data
        """
        try:
            usr_object = Usrbin(self.filename)

            # loop over all detectors (pages) in USRBIN object
            for det_no, detector in enumerate(usr_object.detector):
                page = Page(estimator=estimator)
                page.title = detector.name

                # USRBIN doesn't support differential binning type, only spatial binning is allowed
                estimator.x = MeshAxis(n=detector.nx,
                                       min_val=detector.xlow,
                                       max_val=detector.xhigh,
                                       name="X",
                                       unit="cm",
                                       binning=MeshAxis.BinningType.linear)
                estimator.y = MeshAxis(n=detector.ny,
                                       min_val=detector.ylow,
                                       max_val=detector.yhigh,
                                       name="Y",
                                       unit="cm",
                                       binning=MeshAxis.BinningType.linear)
                estimator.z = MeshAxis(n=detector.nz,
                                       min_val=detector.zlow,
                                       max_val=detector.zhigh,
                                       name="Z",
                                       unit="cm",
                                       binning=MeshAxis.BinningType.linear)

                page.name = "scorer {}".format(detector.score)
                page.unit = ""

                # unpack detector data
                # TODO cross-check if reshaping is needed
                page.data_raw = np.array(unpackArray(usr_object.readData(det_no)))
                page.error_raw = np.empty_like(page.data_raw)

                estimator.add_page(page)

            return usr_object
        except IOError:
            return None

    def parse_usrbdx(self, estimator):
        """
        USRBDX defines a detector for a boundary crossing fluence or current estimator
        :param estimator: an Estimator object, will be modified here and filled with data
        """
        try:
            usr_object = Usrbdx(self.filename)

            # loop over all detectors (pages) in USRBDX object
            for det_no, detector in enumerate(usr_object.detector):
                page = Page(estimator=estimator)
                page.title = detector.name
                page.area = detector.area  # area of the detector in cm**2

                if detector.nb == 1:
                    energy_binning = MeshAxis.BinningType.linear
                    angle_binning = MeshAxis.BinningType.linear
                elif detector.nb == -1:
                    energy_binning = MeshAxis.BinningType.logarithmic
                    angle_binning = MeshAxis.BinningType.linear
                elif detector.nb == 2:
                    energy_binning = MeshAxis.BinningType.linear
                    angle_binning = MeshAxis.BinningType.logarithmic
                elif detector.nb == -2:
                    energy_binning = MeshAxis.BinningType.logarithmic
                    angle_binning = MeshAxis.BinningType.logarithmic
                else:
                    return Exception("Invalid binning type")

                # USRBDX doesn't support spatial (XYZ) binning type
                # USRBDX provides double differential binning, first axis is kinetic energy (in GeV)
                page.diff_axis1 = MeshAxis(n=detector.ne,  # number of energy intervals for scoring
                                           min_val=detector.elow,  # minimum kinetic energy for scoring (GeV)
                                           max_val=detector.ehigh,  # maximum kinetic energy for scoring (GeV)
                                           name="kinetic energy",
                                           unit="GeV",
                                           binning=energy_binning)

                # second axis is solid angle (in steradians)
                page.diff_axis2 = MeshAxis(n=detector.na,  # number of angular bins
                                           min_val=detector.alow,  # minimum solid angle for scoring
                                           max_val=detector.ahigh,  # maximum solid angle for scoring
                                           name="solid angle",
                                           unit="sr",
                                           binning=angle_binning)

                # detector.fluence corresponds to i2 in WHAT(1) in first card of USBDX
                if detector.fluence == 1:
                    page.name = "fluence"
                elif detector.fluence == 0:
                    page.name = "current"
                else:
                    page.name = ""
                page.unit = "cm-2 GeV-1 sr-1"

                # TODO If the generalised particle is 208.0 (ENERGY) or 211.0 (EM-ENRGY),
                #             the quantity scored is differential energy fluence (if
                #             cosine-weighted) or differential energy current (energy crossing the
                #             surface). In both cases the quantity will be expressed in GeV per
                #             cm2 per energy unit per steradian per primary

                # unpack detector data
                # TODO cross-check if reshaping is needed
                page.data_raw = np.array(unpackArray(usr_object.readData(det_no)))
                page.error_raw = np.empty_like(page.data_raw)

                estimator.add_page(page)

            return usr_object
        except IOError:
            return None

    def parse_usrtrack(self, estimator):
        """
        :param estimator: an Estimator object, will be modified here and filled with data
        USRTRACK defines a detector for a track-length fluence estimator
        """
        try:
            usr_object = UsrTrack(self.filename)

            # loop over all detectors (pages) in USRTRACK object
            for det_no, detector in enumerate(usr_object.detector):
                page = Page(estimator=estimator)
                page.title = detector.name
                page.volume = detector.volume   # volume of the detector in cm**3

                # USRTRACK doesn't support spatial (XYZ) binning type
                if detector.type == 1:
                    energy_binning = MeshAxis.BinningType.linear
                elif detector.type == -1:
                    energy_binning = MeshAxis.BinningType.logarithmic
                else:
                    return Exception("Invalid binning type")

                # USRTRACK provides single differential binning, with diff axis in kinetic energy (in GeV)
                page.diff_axis1 = MeshAxis(n=detector.ne,  # number of energy intervals for scoring
                                           min_val=detector.elow,  # minimum kinetic energy for scoring (GeV)
                                           max_val=detector.ehigh,  # maximum kinetic energy for scoring (GeV)
                                           name="kinetic energy",
                                           unit="GeV",
                                           binning=energy_binning)

                page.name = "fluence"
                page.unit = "cm-2 GeV-1"

                # TODO IMPORTANT! The results of USRTRACK are always given as DIFFERENTIAL
                #             distributions of fluence (or tracklength, if the detector region
                #             volume is not specified) in energy, in units of cm-2 GeV-1 (or
                #             cm GeV-1) per incident primary unit weight. Thus, for example, when
                #             requesting a fluence energy spectrum, to obtain INTEGRAL BINNED
                #             results (fluence in cm-2 or tracklength in cm PER ENERGY BIN per
                #             primary) one must multiply the value of each energy bin by the width
                #             of the bin (even for logarithmic binning)

                # TODO If the generalised particle is 208 (ENERGY) or 211 (EM-ENRGY), the
                #             quantity scored is differential energy fluence (or tracklength, if
                #             the detector region volume is not specified), expressed in GeV per
                #             cm2 (or cm GeV) per energy unit per primary. That can sometimes lead
                #             to confusion since GeV cm-2 GeV-1 = cm-2, where energy does not appear.
                #             Note that integrating over energy one gets GeV/cm2.

                # unpack detector data
                # TODO cross-check if reshaping is needed
                page.data_raw = np.array(unpackArray(usr_object.readData(det_no)))
                page.error_raw = np.empty_like(page.data_raw)

                estimator.add_page(page)
            return usr_object
        except IOError:
            return None

    def parse_resnuclei(self, estimator):
        """
        TODO add support for resnuclei
        RESNUCLEi Scores residual nuclei produced in inelastic interactions on a region basis
        :param estimator: an Estimator object, will be modified here and filled with data
        """
        try:
            usr_object = Resnuclei(self.filename)
            # loop over all detectors (pages) in USRTRACK object
            for det_no, detector in enumerate(usr_object.detector):
                page = Page(estimator=estimator)

                # unpack detector data
                # TODO cross-check if reshaping is needed
                page.data_raw = np.array(unpackArray(usr_object.readData(det_no)))
                page.error_raw = np.empty_like(page.data_raw)

                estimator.add_page(page)
            return usr_object
        except IOError:
            return None

    def parse_data(self, estimator):
        """
        TODO
        :param estimator: an Estimator object, will be modified here and filled with data
        """
        for parse_function in (self.parse_usrbin, self.parse_usrtrack, self.parse_usrbdx, self.parse_resnuclei):
            usr_object = parse_function(estimator)

            # stop on first method which return not-None results and return it back
            if usr_object:
                return usr_object

    def read_data(self, estimator, nscale=1):
        """
        TODO
        :param estimator: an Estimator object, will be modified here and filled with data
        """
        usr_object = self.parse_data(estimator)

        estimator.file_counter = usr_object.nbatch
        estimator.number_of_primaries = usr_object.ncase
        estimator.time = usr_object.time
        estimator.title = usr_object.title

        estimator.file_format = 'fluka_binary'
        return True



from copy import deepcopy
import logging
import os
import shutil
import subprocess
import timeit
from multiprocessing import Pool

from enum import IntEnum

from pymchelper.input_output import frompattern


class OutputDataType(IntEnum):
    """
    Output type requested by user plots (i.e .png) or text data
    """
    plot = 1
    txt = 2


class Runner:
    """
    Main class responsible for configuring and starting multiple parallel MC simulation processes
    It can be used to access combined averaged results of the simulation.
    """
    def __init__(self, jobs=None, keep_workspace_after_run=False, output_directory='.'):

        # create pool of processes, waiting to be started by run method
        # if jobs is not specified, os.cpu_count() would be used
        self._pool = Pool(processes=jobs)

        # User of the runner has two options: either to specify number of parallel jobs by
        # setting the self.jobs to given number, or to leave it as None. If self.jobs is None
        # then multiprocessing library will automatically allocate number of parallel jobs to `os.cpu_count()`
        # Therefore we cannot rely of self.jobs as the actual counter of parallel processes
        # Instead we use undocumented feature of multiprocessing module,
        # extracting actual number of allocated processes from `_processes` attribute in Pool class
        # To see how it is used internally in the Python (in v3.9) source code take a look at:
        # https://github.com/python/cpython/blob/3.9/Lib/multiprocessing/pool.py#L210
        self.jobs = self._pool._processes

        # workspace is a collection of working directories
        # this manager is responsible for creating and cleaning working directories
        self.workspace_manager = WorkspaceManager(output_directory=output_directory,
                                                  keep_workspace_after_run=keep_workspace_after_run)

    def run(self, settings):
        """
        Execute parallel simulation processes, creating workspace (and working directories) in the `output_directory`
        In case of successful execution return True, otherwise return False
        """
        start_time = timeit.default_timer()

        # SHIELD-HIT12A and Fluka require RNG seeds to be integers greater or equal to 1
        # each of the workers needs to have its own different RNG seed
        rng_seeds = range(1, self.jobs + 1)

        # create working directories
        self.workspace_manager.create_working_directories(simulation_input_path=settings.input_path,
                                                          rng_seeds=rng_seeds)

        # rng seeds injection to settings for each SingleSimulationExecutor call
        # TODO consider better way of doing it  # skipcq: PYL-W0511
        settings_list = []
        for rng_seed in rng_seeds:
            current_settings = deepcopy(settings)  # do not modify original arguments
            current_settings.set_rng_seed(rng_seed)
            settings_list.append(current_settings)

        # create executor callable object for current run
        executor = SingleSimulationExecutor()

        try:
            # start execution using pool of workers, mapping the executor callable object to the different settings
            # and workspaces
            self._pool.map(executor, zip(settings_list, self.workspace_manager.working_directories_abs_paths))
        except KeyboardInterrupt:
            logging.info('Terminating the pool')
            self._pool.terminate()
            logging.info('Pool is terminated')
            self.workspace_manager.clean()
            return False

        elapsed = timeit.default_timer() - start_time
        logging.info("run elapsed time {:.3f} seconds".format(elapsed))
        return True

    def get_data(self):
        """
        Scans the output directory for location of the working directories (like run_1, run_2).
        Takes all files from all working directories in `output_dir`,
        merges their content to form pymchelper Estimator objects.
        For each of the output file a single Estimator objects is created, which holds numpy arrays with results.
        Return dictionary with keys being output filenames, and values being Estimator objects
        """
        start_time = timeit.default_timer()

        # TODO line below is specific to SHIELD-HIT12A, should be generalised  # skipcq: PYL-W0511
        # scans output directory for MC simulation output files
        output_files_pattern = os.path.join(self.workspace_manager.output_dir_absolute_path, "run_*", "*.bdo")
        logging.debug("Files to merge {:s}".format(output_files_pattern))

        estimators_dict = {}
        # convert output files to list of estimator objects
        estimators_list = frompattern(output_files_pattern)
        for estimator in estimators_list:
            logging.debug("Appending estimator for {:s}".format(estimator.file_corename))
            estimators_dict[estimator.file_corename] = estimator
        elapsed = timeit.default_timer() - start_time
        logging.info("Workspace reading {:.3f} seconds".format(elapsed))

        return estimators_dict

    def clean(self):
        """
        Removes all working directories (if exists)
        """
        self.workspace_manager.clean()


class SingleSimulationExecutor:
    """
    Callable class responsible for execution of the single MC simulation process.
    """

    def __call__(self, settings_and_working_dir, **kwargs):

        # we deliberately combine settings and list of working directories
        # in the single argument `settings_and_working_dir`
        # as this would simplify using this class by multiprocessing module
        settings, working_dir_abs_path = settings_and_working_dir
        try:
            # combine MC engine executable with its command line options to form core of the command string
            # this will form basis of the command, like:
            # /usr/local/bin/shieldhit --time 00:30:50 -v -N 3
            core_command_string = str(settings)

            # for easier digesting by subprocess module, convert command string to a list
            # and finally append the location of the input files
            # finally we obtain a list like:
            # ('/usr/local/bin/shieldhit', '--time', '00:30:50', '-v', '-N', '3', '/data/my/simulation/input')
            command_as_list = core_command_string.split()
            command_as_list.append(working_dir_abs_path)

            # execute the MC simulation on a spawned process
            # TODO handle this differently, i.e. redirect it to file or save in some variable   # skipcq: PYL-W0511
            logging.debug('working directory {:s}, command {:s}'.format(working_dir_abs_path,
                                                                        ' '.join(command_as_list)))
            DEVNULL = open(os.devnull, 'wb')
            subprocess.check_call(command_as_list, cwd=working_dir_abs_path, stdout=DEVNULL, stderr=DEVNULL)
        except KeyboardInterrupt:
            logging.debug("KeyboardInterrupt")
            raise KeyboardInterrupt


class WorkspaceManager:
    """
    A workspace consists of multiple working directories (i.e. run_1, run_2),
    each per one of the parallel simulation run.
    """
    def __init__(self, output_directory='.', keep_workspace_after_run=False):
        self.output_dir_absolute_path = os.path.abspath(output_directory)
        self.keep_workspace_after_run = keep_workspace_after_run
        self.working_directories_abs_paths = []

    def create_working_directories(self, simulation_input_path, rng_seeds=()):
        """
        Create working directories and fill `self.working_directories_abs_paths`
        """
        self.working_directories_abs_paths = []
        for rng_seed in rng_seeds:
            # set workspace to a subdirectory of the output_directory, with run_* pattern
            # i.e. /home/user/output/run_3
            working_dir_abs_path = os.path.join(self.output_dir_absolute_path, 'run_{:d}'.format(rng_seed))
            logging.info("Workspace {:s}".format(working_dir_abs_path))

            if os.path.isdir(simulation_input_path):
                # if path already exists, remove it before copying with copytree()
                if os.path.exists(working_dir_abs_path):
                    shutil.rmtree(working_dir_abs_path)
                # if cleaned or not existing, then create it
                if not os.path.exists(working_dir_abs_path):
                    os.makedirs(working_dir_abs_path)
                # copy all files from the directory
                for directory_entry in os.listdir(simulation_input_path):
                    path_to_directory_entry = os.path.join(simulation_input_path, directory_entry)
                    if os.path.isfile(path_to_directory_entry):
                        shutil.copy2(path_to_directory_entry, working_dir_abs_path)
                logging.debug("Copying input files into {:s}".format(working_dir_abs_path))
            elif os.path.isfile(simulation_input_path):
                if not os.path.exists(working_dir_abs_path):
                    os.makedirs(working_dir_abs_path)
                shutil.copy2(simulation_input_path, working_dir_abs_path)
                logging.debug("Copying input files into {:s}".format(working_dir_abs_path))
            else:
                logging.debug("Input files {:s} not a dir or file".format(simulation_input_path))

            self.working_directories_abs_paths.append(working_dir_abs_path)

    def clean(self):
        """
        clean the workspace by removing all working directories
        (only if requested by `keep_workspace_after_run` flag)
        """
        if not self.keep_workspace_after_run:
            start_time = timeit.default_timer()
            for working_dir_abs_path in self.working_directories_abs_paths:
                # shutil.rmtree will throw exception if the directory we are trying to remove doesn't exist
                # hence we only remove existing directories
                # this allows safely to call `clean` method multiple times
                if os.path.exists(working_dir_abs_path):
                    shutil.rmtree(working_dir_abs_path)
            elapsed = timeit.default_timer() - start_time
            print("Cleaning {:.3f} seconds".format(elapsed))



"""
Module for reading DICOM and PLD files.

One plan may contain one or more fields.
One field may contain one or more layers.
One layer may contain one or more spots.
"""

import sys
import logging
import argparse
import pymchelper

import pydicom as dicom
import numpy as np
from pathlib import Path
from typing import Optional

from dataclasses import dataclass, field
from math import exp, log
from scipy.interpolate import interp1d

logger = logging.getLogger(__name__)

s2fwhm = 2.0 * np.sqrt(2.0 * np.log(2.0))  # 1 FWHM = 2.355 * sigma


def dedx_air(energy: float) -> float:
    """
    Calculate the mass stopping power of protons in air following ICRU 49.

    Valid from 1 to 500 MeV only.

    :params energy: Proton energy in MeV
    :returns: mass stopping power in MeV cm2/g
    """
    if energy > 500.0 or energy < 1.0:
        logger.error("Proton energy must be between 1 and 500 MeV.")
        raise ValueError(f"Energy = {energy:.2f} out of bounds.")
    x = log(energy)
    y = 5.4041 - 0.66877 * x - 0.034441 * (x**2) - 0.0010707 * (x**3) + 0.00082584 * (x**4)
    return exp(y)


class BeamModel():
    """Beam model from a given CSV file."""

    def __init__(self, fn: Path, nominal=True):
        """
        Load a beam model given as a CSV file.

        Interpolation lookup can be done as a function of nominal energy (default, nominal=True),
        or as a function of actual energy (nominal=False).

        Header rows will be discarded and must be prefixed with '#'.

        Input columns for beam model:
            1) nominal energy [MeV]
            2) measured energy [MeV]
            3) energy spread 1 sigma [MeV]
            4) primary protons per MU [protons/MU]
            5) 1 sigma spot size x [mm]
            6) 1 sigma spot size y [mm]
        Optionally, 4 more columns may be given:
            7) 1 sigma divergence x [rad]
            8) 1 sigma divergence y [rad]
            9) cov (x, x') [mm]
            10) cov (y, y') [mm]

        TODO: get rid of scipy dependency
        """
        data = np.genfromtxt(fn, delimiter=",", invalid_raise=False, comments='#')

        # resolve by nominal energy
        if nominal:
            energy = data[:, 0]
        else:
            energy = data[:, 1]

        k = 'cubic'

        cols = len(data[0])
        logger.debug("Number of columns in beam model: %i", cols)

        self.has_divergence = False

        if cols in (6, 10):
            self.f_en = interp1d(energy, data[:, 0], kind=k)  # nominal energy [MeV]
            self.f_e = interp1d(energy, data[:, 1], kind=k)  # measured energy [MeV]
            self.f_espread = interp1d(energy, data[:, 2], kind=k)  # energy spread 1 sigma [% of measured energy]
            self.f_ppmu = interp1d(energy, data[:, 3], kind=k)  # 1e6 protons per MU  [1e6/MU]
            self.f_sx = interp1d(energy, data[:, 4], kind=k)  # 1 sigma x [cm]
            self.f_sy = interp1d(energy, data[:, 5], kind=k)  # 1 sigma y [cm]
        else:
            logger.error("invalid column count")

        if cols == 10:
            logger.debug("Beam model has divergence data")
            self.has_divergence = True
            self.f_divx = interp1d(energy, data[:, 6], kind=k)  # div x [rad]
            self.f_divy = interp1d(energy, data[:, 7], kind=k)  # div y [rad]
            self.f_covx = interp1d(energy, data[:, 8], kind=k)  # cov (x, x') [mm]
            self.f_covy = interp1d(energy, data[:, 9], kind=k)  # cov (y, y') [mm]

        self.data = data


@dataclass
class Layer:
    """
    Handle layers in a plan.

    spots : np.array([[x_i, y_i, mu_i, n], [...], ...) for i spots.
            x,y : are spot positions at isocenter in [mm].
            mu  : are monitor units or meterset weights for the individual spots [MU]
            n   : is the estimated number of primary particles for this spot
    spotsize: np.array() FWHM width of spot in along x and y axis, respectively [mm]
    enorm : nominal energy in [MeV]
    emeas : measured energy in [MeV] at exit nozzle
    cum_mu : cumulative monitor units for this layers [MU]
    repaint : number of repainting, 0 for no repaints TODO: check what is convention here.
    n_spots : number of spots in total
    mu_to_part_coef : conversion coefficient from MU to number of particles (depends on energy)
    """

    spots: np.array = field(default_factory=np.array)
    spotsize: np.array = field(default_factory=np.array)
    energy_nominal: float = 100.0
    energy_measured: float = 100.0
    espread: float = 0.0
    cum_mu: float = 0.0
    cum_particles: float = 0.0  # cumulative number of particles
    xmin: float = 0.0
    xmax: float = 0.0
    ymin: float = 0.0
    ymax: float = 0.0
    repaint: int = 0
    n_spots: int = 1
    mu_to_part_coef: float = 1.0
    is_empty: bool = True  # marker if there are no MUs in this layer


@dataclass
class Field:
    """A single field."""

    layers: list = field(default_factory=list)  # https://stackoverflow.com/questions/53632152/
    n_layers: int = 0  # number of layers in this field
    dose: float = 0.0  # dose in [Gy]
    cum_mu: float = 0.0  # cumulative MU of all layers in this field
    cum_particles: float = 0.0  # cumulative number of particles
    pld_csetweight: float = 0.0  # IBA specific
    # gantry: float = 0.0
    # couch: float = 0.0
    scaling: float = 1.0  # scaling applied to all particle numbers
    xmin: float = 0.0
    xmax: float = 0.0
    ymin: float = 0.0
    ymax: float = 0.0

    def diagnose(self):
        """Print overview of field."""
        energy_list = [layer.energy_nominal for layer in self.layers]
        emin = min(energy_list)
        emax = max(energy_list)

        indent = "   "  # indent layer output, since this is a branch

        print(indent + "------------------------------------------------")
        print(indent + f"Energy layers          : {self.n_layers:10d}")
        print(indent + f"Total MUs              : {self.cum_mu:10.4f}")
        print(indent + f"Total particles        : {self.cum_particles:10.4e} (estimated)")
        print(indent + "------------------------------------------------")
        for i, layer in enumerate(self.layers):
            print(indent + f"   Layer {i: 3}: {layer.energy_nominal: 10.4f} MeV " + f"   {layer.n_spots:10d} spots")
        print(indent + "------------------------------------------------")
        print(indent + f"Highest energy         : {emin:10.4f} MeV")
        print(indent + f"Lowest energy          : {emax:10.4f} MeV")
        print(indent + "------------------------------------------------")
        print(indent + f"Spot field min/max X   : {self.xmin:+10.4f} {self.xmax:+10.4f} mm")
        print(indent + f"Spot field min/max Y   : {self.ymin:+10.4f} {self.ymax:+10.4f} mm")
        print(indent + "------------------------------------------------")
        print("")


@dataclass
class Plan:
    """
    Class for handling treatment plans.

    One plan may consist of one or more fields.
    One field may contain one of more layers.

    Beam model is optional, but needed for exact modeling of the beam.
    If no beam model is given, MUs are translated to particle numbers using approximate stopping power for air (dEdx)
    and empirical scaling factors.
    """

    fields: list = field(default_factory=list)  # https://stackoverflow.com/questions/53632152/
    patient_id: str = ""  # ID of patient
    patient_name: str = ""  # Last name of patient
    patient_initals: str = ""  # Initials of patient
    patient_firstname: str = ""  # Last name of patient
    plan_label: str = ""  #
    plan_date: str = ""  #
    n_fields: int = 0
    beam_model: Optional[BeamModel] = None  # optional beam model class
    beam_name: str = ""
    flip_xy: bool = False  # flag whether x and y has been flipped

    # factor holds the number of particles * dE/dx / MU = some constant
    # MU definitions is arbitrary and my vary from vendor to vendor.
    # This will only be used if no beam model is available, and is based on estimates.
    factor: float = 1.0  # vendor specific factor needed for translating MUs to number of particles
    scaling: float = 1.0

    def apply_beammodel(self):
        """Adjust plan to beam model."""
        if self.beam_model:
            for myfield in self.fields:
                for layer in myfield.layers:
                    # calculate number of particles
                    layer.mu_to_part_coef = self.beam_model.f_ppmu(layer.energy_nominal)
                    layer.energy_measured = self.beam_model.f_e(layer.energy_nominal)
                    layer.espread = self.beam_model.f_espread(layer.energy_nominal)
                    layer.spots[:, 3] = layer.spots[:, 2] * layer.mu_to_part_coef * myfield.scaling
                    layer.spotsize = np.array(
                        [self.beam_model.f_sx(layer.energy_nominal),
                         self.beam_model.f_sy(layer.energy_nominal)]) * s2fwhm
        else:
            for myfield in self.fields:
                for layer in myfield.layers:
                    # if there is no beam model available, we will simply use air stopping power
                    # since MU is proportional to dose in monitor chamber, which means fluence ~ D_air / dEdx(air)
                    layer.mu_to_part_coef = self.factor / dedx_air(layer.energy_measured)
                    layer.spots[:, 3] = layer.spots[:, 2] * layer.mu_to_part_coef * myfield.scaling
                    # old IBA code something like:
                    # weight = mu_to_part_coef * _mu2 * field.cmu / field.pld_csetweight
                    # phi_weight = weight / dedx_air(layer.energy_measured)

        # set cumulative sums
        for myfield in self.fields:
            myfield.cum_particles = 0.0
            myfield.cum_mu = 0.0
            myfield.xmin = 0.0
            myfield.xmax = 0.0
            myfield.ymin = 0.0
            myfield.ymax = 0.0

            # set layer specific values
            for layer in myfield.layers:
                layer.n_spots = len(layer.spots)
                layer.xmin = layer.spots[:, 0].min()
                layer.xmax = layer.spots[:, 0].max()
                layer.ymin = layer.spots[:, 1].min()
                layer.ymax = layer.spots[:, 1].max()
                layer.cum_mu = layer.spots[:, 2].sum()
                if layer.cum_mu > 0.0:
                    layer.is_empty = False
                layer.cum_particles = layer.spots[:, 3].sum()

                myfield.cum_particles += layer.cum_particles
                myfield.cum_mu += layer.cum_mu

                if layer.xmin < myfield.xmin:
                    myfield.xmin = layer.xmin
                if layer.xmax > myfield.xmax:
                    myfield.xmax = layer.xmax

                if layer.ymin < myfield.ymin:
                    myfield.ymin = layer.ymin
                if layer.ymax > myfield.ymax:
                    myfield.ymax = layer.ymax

    def diagnose(self):
        """Print overview of plan."""
        print("Diagnostics:")
        print("---------------------------------------------------")
        print(f"Patient Name           : '{self.patient_name}'       [{self.patient_initals}]")
        print(f"Patient ID             : {self.patient_id}")
        print(f"Plan label             : {self.plan_label}")
        print(f"Plan date              : {self.plan_date}")
        print(f"Number of Fields       : {self.n_fields:2d}")

        for i, myfield in enumerate(self.fields):
            print("---------------------------------------------------")
            print(f"   Field                  : {i + 1:02d}/{self.n_fields:02d}:")
            myfield.diagnose()
            print("")

    def export(self, fn: Path, cols: int, field_nr: int, nominal: bool):
        """
        Export file to sobp.dat format, 'cols' marking the number of columns.

        fn : filename
        cols : number of columns for output format.
               5 column format: energy[GeV] x[cm] y[cm] FWHM[cm] weight
               6 column format: energy[GeV] x[cm] y[cm] FWHMx[cm] FWHMy[cm] weight
               7 column format: energy[GeV] sigmaT0[GeV] x[cm] y[cm] FWHM[cm] weight
        field_nr: in case of multiple field, select what field to export, use '0' to export all fields.
        nominal : flag whether norminal energy should be exported or not

        TODO:
               11 columns: ENERGY, ESPREAD, X, Y, FWHMx, FWHMy, WEIGHT, DIVx, DIVy, COVx, COVy
        """
        if cols == 7:
            header = "*ENERGY(GEV) SigmaT0(GEV) X(CM)   Y(CM)    FWHMx(cm) FWHMy(cm) WEIGHT\n"
        elif cols == 6:
            header = "*ENERGY(GEV) X(CM)   Y(CM)    FWHMx(cm) FWHMy(cm) WEIGHT\n"
        elif cols == 5:
            header = "*ENERGY(GEV) X(CM)   Y(CM)    FWHMx(cm) FWHMy(cm) WEIGHT\n"
        else:
            raise ValueError(f"Output format with {cols} columns is not supported.")

        for i, myfield in enumerate(self.fields):
            j = i + 1  # j is the field number, i is the field index.
            output = header
            # in case all fields should be written to disk, build a new filename with _XX added to the stem
            # of the filename, based on the field number (not the filed index), i.e. first file is sobp_01.dat
            if field_nr == 0:
                fout = Path(fn.parent, (f"{fn.stem}_{j:02d}{fn.suffix}"))
            else:
                # otherwise, check if this is a field the user wanted, if not, skip it.
                fout = fn
                if (j) != field_nr:
                    continue

            for layer in myfield.layers:

                # DICOM and PLD sometimes have empty layers. This will be skipped, to not clutter the sobp.dat file.
                if layer.is_empty:
                    continue

                # Do some conversions, since sobp.dat hold different units.
                if nominal:
                    energy = layer.energy_nominal * 0.001  # convert MeV -> GeV
                else:
                    energy = layer.energy_measured * 0.001  # convert MeV -> GeV
                espread = layer.espread * 0.001  # convert MeV -> GeV

                # Check if field-flip was requested. Then do so for FWHMxy and spot positions
                if self.flip_xy:
                    fwhmy, fwhmx = layer.spotsize * 0.1  # mm -> cm
                else:
                    fwhmx, fwhmy = layer.spotsize * 0.1  # mm -> cm

                for spot in layer.spots:
                    if self.flip_xy:
                        xpos = spot[1] * 0.1  # mm -> cm
                        ypos = spot[0] * 0.1  # mm -> cm
                    else:
                        xpos = spot[0] * 0.1  # mm -> cm
                        ypos = spot[1] * 0.1  # mm -> cm

                    wt = spot[3]
                    # format output file. Carefully tuned so they appear in nice columns synced to header. Maybe.
                    if cols == 7:
                        s = f"{energy:8.6f}     {espread:10.8f}  " \
                            + f"{xpos:6.2f}   {ypos:6.2f}  " \
                            + f"{fwhmx:6.2f}   {fwhmy:6.2f}     {wt:10.4e}\n"
                    elif cols == 6:
                        s = f"{energy:8.6f}     " \
                            + f"{xpos:6.2f}   {ypos:6.2f}  " \
                            + f"{fwhmx:6.2f}   {fwhmy:6.2f}     {wt:10.4e}\n"
                    else:
                        s = f"{energy:8.6f}     " \
                            + f"{xpos:6.2f}   {ypos:6.2f}  " \
                            + f"{fwhmx:6.2f}   {wt:10.4e}\n"
                    output += s
            logger.debug("Export field %d %s, %g MeV", j, fout, myfield.layers[0].energy_nominal)
            fout.write_text(output)  # still in field loop, output for every field


def load(file: Path, beam_model: BeamModel, scaling: float, flip_xy: bool) -> Plan:
    """Load file, autodiscovery by suffix."""
    logger.debug("load() autodiscovery %s", file)
    ext = file.suffix.lower()  # extract suffix, incl. dot separator

    if ext == ".pld":
        logger.debug("autodiscovery: Found a IBA pld file.")
        p = load_PLD_IBA(file, scaling)
    elif ext == ".dcm":
        logger.debug("autodiscovery: Found a DICOM file.")
        p = load_DICOM_VARIAN(file, scaling)  # so far I have no other dicom files
    elif ext == ".rst":
        logger.debug("autodiscovery: Found a GSI raster scan file.")
        p = load_RASTER_GSI(file, scaling)
    else:
        raise ValueError(f"autodiscovery: Unknown file type. {file}")

    # apply beam model if available
    if beam_model:
        p.beam_model = beam_model
    else:
        logger.debug("BeamModel is unavailable in Plan.")

    p.apply_beammodel()
    p.flip_xy = flip_xy

    return p


def load_PLD_IBA(file_pld: Path, scaling=1.0) -> Plan:
    """
    Load a IBA-style PLD-file.

    file_pld : a file pointer to a .pld file, opened for reading.
    Here we assume there is only a single field in every .pld file.
    """
    eps = 1.0e-10

    current_plan = Plan()

    myfield = Field()  # avoid collision with dataclasses.field
    current_plan.fields = [myfield]
    current_plan.n_fields = 1

    # p.factor holds the number of particles * dE/dx / MU = some constant
    # p.factor = 8.106687e7  # Calculated Nov. 2016 from Brita's 32 Gy plan. (no dE/dx)
    current_plan.factor = 5.1821e8  # protons per (MU/dEdx), Estimated calculation Apr. 2017 from Brita's 32 Gy plan.

    # currently scaling is treated equal at plan and field level. This is for future use.
    current_plan.scaling = scaling
    field.scaling = scaling

    pldlines = file_pld.read_text().split('\n')
    pldlen = len(pldlines)
    logger.info("Read %d lines of data.", pldlen)

    field.layers = []
    field.n_layers = 0

    # First line in PLD file contains both plan and field data
    tokens = pldlines[0].split(",")
    current_plan.patient_id = tokens[1].strip()
    current_plan.patient_name = tokens[2].strip()
    current_plan.patient_initals = tokens[3].strip()
    current_plan.patient_firstname = tokens[4].strip()
    current_plan.plan_label = tokens[5].strip()
    current_plan.beam_name = tokens[6].strip()
    field.cmu = float(tokens[7].strip())  # total amount of MUs in this field
    field.pld_csetweight = float(tokens[8].strip())
    field.n_layers = int(tokens[9].strip())  # number of layers

    espread = 0.0  # will be set by beam model

    for i in range(1, pldlen):  # loop over all lines starting from the second one
        line = pldlines[i]
        if "Layer" in line:  # each new layers starts with the "Layer" keyword
            # the "Layer" header is formated as
            # "Layer, "
            header = line
            tokens = header.split(",")
            # extract the subsequent lines with elements
            el_first = i + 1
            el_last = el_first + int(tokens[4])

            elements = pldlines[el_first:el_last]  # each line starting with "Element" string is a spot.

            # tokens[0] just holds the "Layer" keyword
            # IBA PLD holds nominal spot size in 1D, 1 sigma in [mm]
            spotsize = float(tokens[1].strip()) * s2fwhm  # convert mm sigma to mm FWHM (this is just a float)

            energy_nominal = float(tokens[2].strip())
            cmu = float(tokens[3].strip())
            nspots = int(tokens[4].strip())
            logger.debug(tokens)

            # read number of repaints only if 5th column is present, otherwise set to 0
            nrepaint = 0  # we suspect repaints = 1 means all dose will be delivered once.
            if len(tokens) > 5:
                nrepaint = int(tokens[5].strip())

            spots = np.array([])

            layer = Layer(spots=spots,
                          spotsize=np.array([spotsize, spotsize]),
                          energy_nominal=energy_nominal,
                          energy_measured=energy_nominal,
                          espread=espread,
                          cum_mu=cmu,
                          n_spots=nspots,
                          repaint=nrepaint)

            for element in elements:  # loop over each spot in this layer
                token = element.split(",")
                # the .pld file has every spot position repeated, but MUs are only in
                # every second line, for reasons unknown.
                _x = float(token[1].strip())
                _y = float(token[2].strip())
                _mu = float(token[3].strip())

                # fix bad float conversions
                if np.abs(_x) < eps:
                    _x = 0.0
                if np.abs(_y) < eps:
                    _y = 0.0
                if _mu < eps:
                    _mu = 0.0

                # PLD files have the spots listed tiwce, once with no MUs. These are removed here.
                if _mu > 0.0:
                    layer.spots = np.append([layer.spots], [_x, _y, _mu, _mu])
                else:
                    # this was an empty spot, decrement spot count, and do not add it.
                    nspots -= 1

            layer.spots = layer.spots.reshape(nspots, 4)
            current_plan.fields[0].layers.append(layer)

            logger.debug("appended layer %i with %i spots", len(current_plan.fields[0].layers), layer.n_spots)
    return current_plan


def load_DICOM_VARIAN(file_dcm: Path, scaling=1.0) -> Plan:
    """Load varian type dicom plans."""
    ds = dicom.dcmread(file_dcm)
    # Total number of energy layers used to produce SOBP

    p = Plan()
    p.patient_id = ds['PatientID'].value
    p.patient_name = ds['PatientName'].value
    p.patient_initals = ""
    p.patient_firstname = ""
    p.plan_label = ds['RTPlanLabel'].value
    p.plan_date = ds['RTPlanDate'].value

    # protons per (MU/dEdx), Estimated calculation Nov. 2022 from DCPT beam model
    p.factor = 17247566.1
    p.scaling = scaling  # nee note in IBA reader above.
    espread = 0.0  # will be set by beam model
    p.n_fields = int(ds['FractionGroupSequence'][0]['NumberOfBeams'].value)
    logger.debug("Found %i fields", p.n_fields)

    dcm_fgs = ds['FractionGroupSequence'][0]['ReferencedBeamSequence']  # fields for given group number

    for i, dcm_field in enumerate(dcm_fgs):
        myfield = Field()
        logger.debug("Appending field number %d...", i)
        p.fields.append(myfield)
        myfield.dose = float(dcm_field['BeamDose'].value)
        myfield.cum_mu = float(dcm_field['BeamMeterset'].value)
        myfield.pld_csetweight = 1.0
        myfield.scaling = scaling  # nee note in IBA reader above.
        myfield.n_layers = int(ds['IonBeamSequence'][i]['NumberOfControlPoints'].value)
        dcm_ibs = ds['IonBeamSequence'][i]['IonControlPointSequence']  # layers for given field number
        logger.debug("Found %i layers in field number %i", myfield.n_layers, i)

        cmu = 0.0

        for j, layer in enumerate(dcm_ibs):

            # gantry and couch angle is stored per energy layer, strangely
            if 'NominalBeamEnergy' in layer:
                energy = float(layer['NominalBeamEnergy'].value)  # Nominal energy in MeV
            if 'NumberOfScanSpotPositions' in layer:
                nspots = int(layer['NumberOfScanSpotPositions'].value)  # number of spots
                logger.debug("Found %i spots in layer number %i at energy %f", nspots, j, energy)
            if 'NumberOfPaintings' in layer:
                nrepaint = int(layer['NumberOfPaintings'].value)  # number of spots
            if 'ScanSpotPositionMap' in layer:
                _pos = np.array(layer['ScanSpotPositionMap'].value).reshape(nspots, 2)  # spot coords in mm
            if 'ScanSpotMetersetWeights' in layer:
                _mu = np.array(layer['ScanSpotMetersetWeights'].value).reshape(nspots, 1)  # spot MUs
                cmu = _mu.sum()
            if 'ScanningSpotSize' in layer:
                # Varian dicom holds nominal spot size in 2D, FWHMMx,y in [mm]
                spotsize = np.array(layer['ScanningSpotSize'].value)
                spots = np.c_[_pos, _mu, _mu]  # weight will be calculated later when beam model is applied
                myfield.layers.append(Layer(spots, spotsize, energy, energy, espread, cmu, nrepaint, nspots))
    return p


def load_RASTER_GSI(file_rst: Path, scaling=1.0):
    """this is implemented in pytrip, maybe we could import it?."""
    logging.warning("GSI raster file reader not implemented yet.")
    logging.info("Opening file %s", file_rst)
    p = Plan()
    p.scaling = scaling  # nee note in IBA reader above.
    return p


def main(args=None) -> int:
    """
    Read a plan file (dicom, pld, rst), and convert it to a spot list, easy to read by MC codes.

    The MU based spot list in dicom/pld/rst is converted to particle weighted spot list,
    optionally based on a realistic beam model, or on simple estimations.
    """
    if args is None:
        args = sys.argv[1:]

    parser = argparse.ArgumentParser()
    parser.add_argument('fin',
                        metavar="input_file",
                        type=Path,
                        help="path to input file in IBA '.pld'-format or Varian DICOM-RN.")
    parser.add_argument('fout',
                        nargs='?',
                        metavar="output_file",
                        type=Path,
                        help="path to the SHIELD-HIT12A/FLUKA output_file. Default: 'sobp.dat'",
                        default="sobp.dat")
    parser.add_argument('-b',
                        metavar="beam_model.csv",
                        type=Path,
                        help="optional input beam model in commasparated CSV format",
                        dest='fbm',
                        default=None)
    parser.add_argument('-i', '--invert', action='store_true', help="invert XY axis", dest="invert", default=False)
    parser.add_argument('-f',
                        '--field',
                        type=int,
                        dest='field_nr',
                        help="select which field to export, for dicom files holding several fields. "
                        + "'0' will produce multiple output files with a running number.",
                        default=1)
    parser.add_argument('-d',
                        '--diag',
                        action='store_true',
                        help="print diagnostics, but do not export data",
                        dest="diag",
                        default=False)
    parser.add_argument('-n',
                        '--nominal',
                        action='store_true',
                        help="save nominal energies instead of beam model energies",
                        dest="nominal",
                        default=False)
    parser.add_argument('-s', '--scale', type=float, dest='scale', help="number of particles*dE/dx per MU", default=1.0)
    parser.add_argument('-c',
                        '--columns',
                        type=int,
                        dest='cols',
                        help="number of columns in output file. 5, 6, 7 col format supported, default is 7.",
                        default=7)
    parser.add_argument('-v', '--verbosity', action='count', help="increase output verbosity", default=0)
    parser.add_argument('-V', '--version', action='version', version=pymchelper.__version__)
    parsed_args = parser.parse_args(args)

    if parsed_args.verbosity == 1:
        logging.basicConfig(level=logging.INFO)

    if parsed_args.verbosity > 1:
        logging.basicConfig(level=logging.DEBUG)

    if parsed_args.fbm:
        bm = BeamModel(parsed_args.fbm)
    else:
        bm = None

    pln = load(parsed_args.fin, bm, parsed_args.scale, parsed_args.invert)

    if parsed_args.diag:
        pln.diagnose()
    else:
        pln.export(parsed_args.fout, parsed_args.cols, parsed_args.field_nr, parsed_args.nominal)

    return 0


if __name__ == '__main__':
    # Run sys.exit with exit code of the main method.
    sys.exit(main(sys.argv[1:]))



# Copyright and User License
# ~~~~~~~~~~~~~~~~~~~~~~~~~~
# Copyright Vasilis.Vlachoudis@cern.ch for the
# European Organization for Nuclear Research (CERN)
#
# All rights not expressly granted under this license are reserved.
#
# Installation, use, reproduction, display of the
# software ("flair"), in source and binary forms, are
# permitted free of charge on a non-exclusive basis for
# internal scientific, non-commercial and non-weapon-related
# use by non-profit organizations only.
#
# For commercial use of the software, please contact the main
# author Vasilis.Vlachoudis@cern.ch for further information.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following
# conditions are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the
#    distribution.
#
# DISCLAIMER
# ~~~~~~~~~~
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT
# NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY, OF
# SATISFACTORY QUALITY, AND FITNESS FOR A PARTICULAR PURPOSE
# OR USE ARE DISCLAIMED. THE COPYRIGHT HOLDERS AND THE
# AUTHORS MAKE NO REPRESENTATION THAT THE SOFTWARE AND
# MODIFICATIONS THEREOF, WILL NOT INFRINGE ANY PATENT,
# COPYRIGHT, TRADE SECRET OR OTHER PROPRIETARY RIGHT.
#
# LIMITATION OF LIABILITY
# ~~~~~~~~~~~~~~~~~~~~~~~
# THE COPYRIGHT HOLDERS AND THE AUTHORS SHALL HAVE NO
# LIABILITY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL,
# CONSEQUENTIAL, EXEMPLARY, OR PUNITIVE DAMAGES OF ANY
# CHARACTER INCLUDING, WITHOUT LIMITATION, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES, LOSS OF USE, DATA OR PROFITS,
# OR BUSINESS INTERRUPTION, HOWEVER CAUSED AND ON ANY THEORY
# OF CONTRACT, WARRANTY, TORT (INCLUDING NEGLIGENCE), PRODUCT
# LIABILITY OR OTHERWISE, ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
# Author:	Vasilis.Vlachoudis@cern.ch
# Date:	18-May-2004

__author__ = "Vasilis Vlachoudis"
__email__ = "Vasilis.Vlachoudis@cern.ch"


class CSGException(Exception):
    pass


MAXEXPR = 10000


def tokenize(expr):
    """return a list of expression tokens"""
    lst = []
    for line in expr.splitlines():
        line = line.strip()
        if len(line) == 0:
            continue
        if line[0] == '*':
            continue  # Skip comments
        line = line.replace("+", " + ")
        line = line.replace("-", " - ")
        line = line.replace("#", " # ")
        line = line.replace("|", " | ")
        line = line.replace("(", " ( ")
        line = line.replace(")", " ) ")
        lst.extend(line.split())
    return lst


def splitZones(expr):
    """split a tokenized expression into zones"""
    zones = []
    zone = []
    depth = 0
    for token in expr:
        if token == "|" and depth == 0:
            if zone:
                zones.append(zone)
                zone = []
            continue

        elif token == "(":
            depth += 1

        elif token == ")":
            depth -= 1
            if depth < 0:
                raise CSGException("Too many closing parenthesis")

        zone.append(token)

    if zone:
        zones.append(zone)
    return zones


def toString(expr):
    """return a string version of a tokenize expression"""
    s = ""
    prev = "("
    for token in expr:
        if prev == "(":
            s += token
        elif token == "|":
            s += " | "
        elif token in ("+", "-"):
            s += " %s" % (token)
        else:
            s += token
        prev = token
    return s


def check(expr):
    """check depth of an rpn expression"""
    depth = 0
    for token in expr:
        if token in ("+", "-", "|"):
            depth -= 1
        else:
            depth += 1
    return depth == 1


def exp2rpn(expr):
    """
    Convert the FLUKA Boolean expression to Reverse Polish Notation (RPN)
    Since the normalization routine does not accept the +/- as signs to
    objects, the routine is converting the leading - to @- (Universe minus)
    where the special symbol @ is treated as the universe.
    Furthermore the leading + is ignored as well as the leading | which is
    accepted by fluka.

    WARNING: No check is done for the correctness of the expression apart
             from the parenthesis nesting

    ie.
           A+B         -> A B +
          (A+B)|C      -> A B + C |
          (A|B)|C+D|E  -> A B | C D + | E |
          -A           -> @ A -
          -(-A)        -> @ @ A - -

    The routine is using the same array for returning the Reverse Polish
    expression, since the format is more compact.
    This is generally true apart one case -A -> @ A -

    Priorities are treated as:
        Operator  Priority         In     Out
        --------  --------        ---     ---
          |       lower             1       2
          +       high              3       4
          -       high              3       4
          (       higher           99       0
          )       higher            0      99
          object  highest         101     100

    Algorithm
     Consider the expression as a train moving on a railroad with a
     T-shape, where each token is one wagon

                         <-  (A|B)|C+D
         ------------.   .------------
         RPN-End      | /      Exp-End
                       |
                      S|
                      t|
                      a|
                      c|
                      k|

     Each wagon to move from the Exp-End to the RPN-End it has to make
     a stop first in the Stack-End. Before entering in the stack, the
     priority-IN will be checked against the objects in the stack.
     All top-most objects currently present in the stack with
     higher priority-OUT will be transfered from the stack to the
     RPN-End. Apart from the opening parenthesis ( which is discarded.

     Example:
     (1)                         A+B|C (1)
     (2) A                        +B|C (2)
     (3) A                         B|C (3)
     (4) A                          |C (4)
     (5) A B +                       C (5)
     (6) A B + C
     (7) A B + C |
         ------------.   .------------
         RPN-End      | /      Exp-End
                       |
                      S|
                      t|
                      a|
                      c|     B   C
                      k| A + + | | |
                         1 2 3 4 5 6 7
    :param expr:
    :return:
    """
    """Replace expression list expr to rpn format"""
    #              Operator In Out
    priorities = {"(": [99, 0], "|": [1, 2], "+": [3, 4], "-": [3, 4], ")": [0, 99], " ": [101, 100]}
    stack = []
    newproduct = 1
    i = 0
    m = 0
    while i < len(expr):
        tcur = expr[i]
        # Check for special leading chars
        if newproduct and tcur in ('|', '+'):
            newproduct = (tcur == '|')
            i += 1
            continue

        if newproduct and tcur == '-':
            expr.insert(i, '*')  # insert space in ith position
            tcur = '@'  # Universe

        newproduct = tcur in ('(', '|')

        # Find priorities
        try:
            prio = priorities[tcur]
        except KeyError:
            prio = priorities[" "]

        ip = prio[0]
        op = prio[1]

        # Remove from the stack everything with higher priority
        while len(stack) > 0 and ip < stack[-1][1]:
            if stack[-1][0] != '(':
                expr[m] = stack.pop()[0]
                m += 1
            else:
                stack.pop()

        # Push it into the stack
        if tcur != ')':
            stack.append([tcur, op])
        else:
            # Should be an opening parenthesis
            if len(stack) == 0:
                raise CSGException("Unbalanced parenthesis")
            stack.pop()

        i += 1

    # Empty Stack
    while len(stack) > 0:
        if stack[-1][0] == '(':
            raise CSGException("Unbalanced parenthesis")

        if m >= len(expr):
            expr.append("")
        expr[m] = stack.pop()[0]
        m = m + 1

    # Delete unwanted items
    for i in range(len(expr) - 1, m - 1, -1):
        del expr[i]


def rpnorm(expr):
    """
    Normalize a CG expression given in Reverse Polish Notation.
    Normalized CG expression is an expression given as sum (Boolean OR) of
    products (Boolean intersection or subtraction).
    The normalization (expansion of parenthesis and operator priorities)
    should be performed by recursively calling the RPNRULE subroutine.
    Since Fortran-77 doesn't have recursion, call the RPNRULE for every
    operator starting from the right-most one, until no rule is found.
    :param expr:
    :return:
    """

    # Loop until there is no any extra change needed
    # Scan to find the first operators
    changed = 1
    while changed:
        changed = 0
        i = len(expr) - 1
        while i >= 4:
            tx = expr[i]
            if tx in ('+', '-', '|'):
                length = len(expr)
                rule = _rpnrule(expr, i)
                if rule > 0:
                    changed = 1
                    i += len(expr) - length + 1
            i -= 1
            if len(expr) > MAXEXPR:
                raise CSGException("Expansion failed. Too many terms")


def _subTerms(expr, n):
    """
    This routine returns the pointers in the RPN terms array of the
    starting point of the left sub-expression LOWLEFT and right
    sub-expression LOWRIGHT given the pointer of the expr-operator NTX.
    The searching is performed by scanning from right to left the number
    of operators and objects pushed into the stack

    EXP          (expr-left) op (expr-right)

    RPN ...... | expr-left | expr-right | op | .......
              Lowleft      LowRight  op-1 ntx
    :param expr:
    :param n:
    :return:
    """
    nop = 0
    lowRight = 0
    lowLeft = 0

    while n >= 0:
        t = expr[n]
        if t in ('+', '-', '|'):
            nop += 1
        else:
            nop -= 1

        n = n - 1
        if nop == 0:
            if lowRight == 0:
                lowRight = n + 1
                nop = nop + 1
                continue
            else:
                lowLeft = n + 1
                return lowLeft, lowRight
    return lowLeft, lowRight


def _rpnrule(expr, n):
    """
    Find a matching rule and apply it on the sub-expression starting from
    the N position in the Reverse Polish Notation

    An expression is in normal form when all the parenthesis are expanded
    and the expression is described as a sum (UNIONS) of products
    (INTERSECTIONS and/or SUBTRACTIONS)

    An expression can be converted to normal form by repeatedly applying
    the following set of production rules to the expression and then to its
    sub-expressions:

       Normal Form                        Reverse Polish Notation
     1. X-(Y|Z) -> (X-Y)-Z                X Y Z | -  ->  X Y - Z -
     2. X+(Y|Z) -> (X+Y)|(X+Z)            X Y Z | +  ->  X Y + X Z + |
     3. X-(Y+Z) -> (X-Y)|(X-Z)            X Y Z + -  ->  X Y - X Z - |
     4. X+(Y+Z) -> (X+Y)+Z                X Y Z + +  ->  X Y + Z +
     5. X-(Y-Z) -> (X-Y)|(X+Z)            X Y Z - -  ->  X Y - X Z + |
     6. X+(Y-Z) -> (X+Y)-Z                X Y Z - +  ->  X Y + Z -
     7. X|(Y|Z) -> (X|Y)|Z                X Y Z | |  ->  X Y | Z |
     8. (X-Y)+Z -> (X+Z)-Y                X Y - Z +  ->  X Z + Y -
     9. (X|Y)-Z -> (X-Z)|(Y-Z)            X Y | Z -  ->  X Z - Y Z - |
    10. (X|Y)+Z -> (X+Z)|(Y+Z)            X Y | Z +  ->  X Z + Y Z + |
    X,Y, and Z here match both primitives or sub-expressions.
    :param expr:
    :param n:
    :return:
    """

    # Reset rule
    rule = 0

    # Top-most operator
    op = expr[n]
    if op not in ('+', '-', '|'):
        return rule

    # Right operator
    rop = expr[n - 1]

    # Find left and right sub-trees
    ll, lr = _subTerms(expr, n)

    # Left operator
    lop = ' '
    if lr > 0:
        lop = expr[lr - 1]

    # Find Rule
    if op == "-" and rop == "|":
        rule = 1
    elif op == "+" and rop == "|":
        rule = 2
    elif op == "-" and rop == "+":
        rule = 3
    elif op == "+" and rop == "+":
        rule = 4
    elif op == "-" and rop == "-":
        rule = 5
    elif op == "+" and rop == "-":
        rule = 6
    elif op == "|" and rop == "|":
        rule = 7
    elif op == "+" and lop == "-":
        rule = 8
    elif op == "-" and lop == "|":
        rule = 9
    elif op == "+" and lop == "|":
        rule = 10
    else:
        return rule

    # Find sub expressions X Y Z
    if rule <= 7:  # X op (Y rop Z)
        Xu = lr - 1
        Xl = ll

        ll, lr = _subTerms(expr, n - 1)
        # Yu = lr - 1    # TODO why not used ?
        # Yl = ll        # TODO why not used ?
        Zu = n - 2
        Zl = lr
    else:  # (X lop Y) op Z
        Zu = n - 1
        Zl = lr
        L = lr - 1
        ll, lr = _subTerms(expr, L)
        Xu = lr - 1
        Xl = ll
        # Yu = L - 1  # TODO why not used ?
        # Yl = lr     # TODO why not used ?

    # Expand the rule
    # 1. X-(Y|Z) -> (X-Y)-Z	 X Y Z | -  ->  X Y - Z -
    if rule == 1:
        # Leave X Y
        # Insert a - operator after Y
        expr.insert(Zl, "-")
        # Chop length by 1
        del expr[Zu + 2]
        # Change the last operator to -
        expr[Zu + 2] = '-'

    # 2. X+(Y|Z) -> (X+Y)|(X+Z)     X Y Z + |  ->  X Y + X Z + |
    elif rule == 2:
        # Leave X Y
        # Insert a + operator after Y
        expr.insert(Zl, "+")
        # Copy X after the + operator
        to = Zl + 1
        expr[to:to] = expr[Xl:Xu + 1]
        Zu += Xu - Xl + 2
        # Change last 2 operators to + |
        expr[Zu + 1] = '+'
        expr[Zu + 2] = '|'

    # 3. X-(Y+Z) -> (X-Y)|(X-Z)     X Y Z + -  ->  X Y - X Z - |
    elif rule == 3:
        # Leave X Y
        # Insert a - operator after Y
        expr.insert(Zl, "-")
        # Copy X after the - operator
        to = Zl + 1
        expr[to:to] = expr[Xl:Xu + 1]
        Zu += Xu - Xl + 2
        # Change last 2 operators to - |
        expr[Zu + 1] = '-'
        expr[Zu + 2] = '|'

    # 4. X+(Y+Z) -> (X+Y)+Z	 X Y Z + +  ->  X Y + Z +
    elif rule == 4:
        # Leave X Y
        # Insert a + operator after Y
        expr.insert(Zl, "+")
        # Chop length by 1
        del expr[Zu + 2]
        # Change the last operator to +
        expr[Zu + 2] = '+'

    # 5. X-(Y-Z) -> (X-Y)|(X+Z)     X Y Z - -  ->  X Y - X Z + |
    elif rule == 5:
        # Leave X Y
        # Insert a - operator after Y
        expr.insert(Zl, "-")
        # Copy X after the - operator
        to = Zl + 1
        expr[to:to] = expr[Xl:Xu + 1]
        Zu += Xu - Xl + 2
        # Change last 2 operators to + |
        expr[Zu + 1] = '+'
        expr[Zu + 2] = '|'

    # 6. X+(Y-Z) -> (X+Y)-Z	 X Y Z - +  ->  X Y + Z -
    elif rule == 6:
        # Leave X Y
        # Insert a + operator after Y
        expr.insert(Zl, "+")
        # Chop length by 1
        del expr[Zu + 2]
        # Change the last operator to -
        expr[Zu + 2] = '-'

    # 7. X|(Y|Z) -> (X|Y)|Z	 X Y Z | |  ->  X Y | Z |
    elif rule == 7:
        # Leave X Y
        # Insert a | operator after Y
        expr.insert(Zl, "|")
        # Chop length by 1
        del expr[Zu + 2]
        # Change the last operator to |
        expr[Zu + 2] = '|'

    # 8. (X-Y)+Z -> (X+Z)-Y	 X Y - Z +  ->  X Z + Y -
    elif rule == 8:
        # Leave X
        # Copy "Z +" after X
        L = Zu - Zl + 2
        to = Xu + 1
        expr[to:to] = expr[Zl:Zl + L]
        # Delete old "Z +"
        del expr[Zl + L:Zl + L + L]

    #  9. (X|Y)-Z -> (X-Z)|(Y-Z)     X Y | Z -  ->  X Z - Y Z - |
    # 10. (X|Y)+Z -> (X+Z)|(Y+Z)     X Y | Z +  ->  X Z + Y Z + |
    elif rule in (9, 10):
        # Leave X
        # Copy "Z -" or "Z +" after X
        L = Zu - Zl + 2
        to = Xu + 1
        expr[to:to] = expr[Zl:Zu + 2]
        # Correct Z position
        Zl += L
        Zu += L
        # Delete the | infront of Z
        del expr[Zl - 1]
        # Add | at the end
        expr.insert(Zu + 1, "|")

    return rule


def rpn2exp(rpn):
    """
    Convert a NORMALIZED Reverse Polish notation to a standard expression
    WARNING: The routine expects an expression where for each UNION
    operator the right-sub-expression is a product while the left can be
    UNION or a product
    :param rpn:
    :return:
    """
    zones = []
    plus = []
    minus = []

    i = 0
    nstack = 0
    endprod = 0
    while i < len(rpn):
        tx = rpn[i]
        if tx in ('+', '-', '|'):
            nstack -= 1
        else:
            # First term is always a plus
            # .. peek then next operator to check for sign
            if len(plus) == 0 or rpn[i + 1] == '+':
                plus.append(tx)
                lastPlus = 1
            elif rpn[i + 1] == '-':
                minus.append(tx)
                lastPlus = 0
            nstack += 1

        if nstack == 0:
            endprod = 1
        elif nstack == 3:
            if lastPlus:
                del plus[-1]
            else:
                del minus[-1]
            i -= 2
            endprod = 1
        elif tx == '|':
            i -= 2
            endprod = 1
        elif i == len(rpn) - 1:
            endprod = 1

        if endprod:
            optZone(plus, minus)
            # remove None terms
            plus = filter(lambda x: x, plus)
            minus = filter(lambda x: x, minus)
            if len(plus) > 0 or len(minus) > 0:
                zones.append((plus, minus))
            plus = []
            minus = []
            nstack = 0
            endprod = 0
        i += 1

    # Remove duplicates of products
    rmDoubles(zones)

    # Reconstruct expression
    expr = []
    for plus, minus in zones:
        if len(expr) > 0 and expr[-1] != "|":
            expr.append("|")
        # Fill the new array
        for j in plus:
            expr.append("+")
            expr.append(j)
        for j in minus:
            expr.append("-")
            expr.append(j)

    return expr


def optZone(plus, minus):
    """
    Optimize a product, by removing the universe @, duplicated terms like
    A+A, A-A, -A-A, and finally calling the geometrical optimizations and
    sorting by name the primitives in both the plus and minus terms.

    The product is described by 2 arrays the PLUS,NPLUS and MINUS,NMINUS
    with all the plus and minus terms of the product

    WARNING: It doesn't delete the term from the arrays but changes the
    name to space.
    :param plus:
    :param minus:
    :return:
    """

    # Remove Universe @ from PLUS
    # and
    # remove duplicate terms A+A=A, A-A={}
    for i in range(len(plus)):
        if plus[i] == '@':
            plus[i] = None

        if plus[i] is not None:
            for j in range(i + 1, len(plus)):
                if plus[i] == plus[j]:
                    plus[j] = None
            for j in range(len(minus)):
                if plus[i] == minus[j]:
                    # Remove everything
                    del plus[:]
                    del minus[:]
                    return

    # Discard product if universe @ exists in MINUS
    # Check for duplicates in MINUS like -A-A=-A
    for i in range(len(minus)):
        if minus[i] == '@':
            # Remove everything
            del plus[:]
            del minus[:]
            return
        if minus[i] is not None:
            for j in range(i + 1, len(minus)):
                if minus[i] == minus[j]:
                    minus[j] = None

    # Perform the Geometrical optimization in the product
    # call OptGeo(nplus,plus,nminus,minus)

    # Peform a bubble sort on product terms
    plus.sort()
    minus.sort()


def rmDoubles(zones):
    """
    Remove duplicates of products A+B|A+B|C+D  ->  A+B|C+D
    :param zones:
    :return:
    """
    i = -1
    while i < len(zones) - 1:
        i += 1
        plus1, minus1 = zones[i]
        j = i
        while j < len(zones) - 1:
            j += 1
            plus2, minus2 = zones[j]

            # Check if they are the same
            if len(plus1) != len(plus2) or len(minus1) != len(minus2):
                continue
            diffplus = -1
            for k in range(len(plus1)):
                if plus1[k] != plus2[k]:
                    diffplus = k
                    break

            diffminus = -1
            for k in range(len(minus1)):
                if minus1[k] != minus2[k]:
                    diffminus = k
                    break

            if diffplus == -1 and diffminus == -1:
                del zones[j]
                j -= 1
                continue


def split(expr):
    """
    Break products into lists
    :param expr:
    :return:
    """
    brk = []
    expr = expr[:]  # Make a copy of the expression
    while 1:
        try:
            p = expr.index("|")
            brk.append(expr[0:p])
            del expr[0:p + 1]
        except Exception:
            brk.append(expr)
            return brk



import logging
import numpy as np

logger = logging.getLogger(__name__)


class TRiP98CubeWriter:
    def __init__(self, filename, options):
        self.output_corename = filename

    def write(self, estimator):
        if len(estimator.pages) > 1:
            print("Conversion of data with multiple pages not supported yet")
            return False

        import getpass
        from pymchelper.shieldhit.detector.detector_type import SHDetType
        from pymchelper import __version__ as _pmcversion
        try:
            from pytrip import __version__ as _ptversion
        except ImportError:
            logger.error("pytrip package missing, to install type `pip install pytrip98`")
            return 1

        pixel_size_x = (estimator.x.max_val - estimator.x.min_val) / estimator.x.n
        pixel_size_z = (estimator.z.max_val - estimator.z.min_val) / estimator.z.n

        logging.debug("psx: {:.6f} [cm]".format(pixel_size_x))
        logging.debug("psz: {:.6f} [cm]".format(pixel_size_z))

        _patient_name = "Anonymous"
        _created_by = getpass.getuser()
        _creation_info = "Created with pymchelper {:s}; using PyTRiP98 {:s}".format(_pmcversion,
                                                                                    _ptversion)

        if estimator.pages[0].dettyp == SHDetType.dose:

            from pytrip import dos

            cube = dos.DosCube()
            # Warning: PyTRiP cube dimensions are in [mm]
            cube.create_empty_cube(
                1.0, estimator.x.n, estimator.y.n, estimator.z.n,
                pixel_size=pixel_size_x * 10.0,
                slice_distance=pixel_size_z * 10.0)

            # .dos dose cubes are usually in normalized integers,
            # where "1000" equals 100.0 % dose.
            # The next are also the defaults, but just to be clear
            # this is specifically set.
            cube.data_type = "integer"
            cube.num_bytes = 2
            cube.pydata_type = np.int16

            cube.cube = estimator.data

            if estimator.tripdose >= 0.0 and estimator.tripntot > 0:
                cube.cube = (cube.cube * estimator.tripntot * 1.602e-10) / estimator.tripdose * 1000.0
            else:
                cube.cube = (cube.cube / cube.cube.max()) * 1200.0

            # Save proper meta information
            cube.patient_name = _patient_name
            cube.created_by = _created_by
            cube.creation_info = _creation_info

            cube.write(self.output_corename)

            return 0

        elif estimator.pages[0].dettyp in (SHDetType.dlet, SHDetType.tlet, SHDetType.dletg, SHDetType.tletg):

            from pytrip import let

            cube = let.LETCube()
            # Warning: PyTRiP cube dimensions are in [mm]
            cube.create_empty_cube(
                1.0, estimator.x.n, estimator.y.n, estimator.z.n,
                pixel_size=pixel_size_x * 10.0,
                slice_distance=pixel_size_z * 10.0)

            # .dosemlet.dos LET cubes are usually in 32 bit floats.
            cube.data_type = "float"
            cube.num_bytes = 4
            cube.pydata_type = np.float32

            # need to redo the cube, since by default np.float32 are allocated.
            # When https://github.com/pytrip/pytrip/issues/35 is fixed,
            # then this should not be needed.
            cube.cube = np.ones((cube.dimz, cube.dimy, cube.dimx), dtype=cube.pydata_type)

            cube.cube = estimator.data
            cube.cube *= 0.1  # MeV/cm -> keV/um
            # Save proper meta information

            cube.patient_name = _patient_name
            cube.created_by = _created_by
            cube.creation_info = _creation_info

            cube.write(self.output_corename)

            return 0

        else:
            logger.error("Tripcube target is only allowed with dose- or LET-type detectors.")
            raise Exception("Illegal detector for tripcube.")



import logging
import os
from enum import IntEnum

import numpy as np

logger = logging.getLogger(__name__)


class PlotAxis(IntEnum):
    x = 1
    y = 2
    z = 3


class PlotDataWriter:
    """plot data writer"""

    def __init__(self, filename, options):
        self.filename = filename
        if not self.filename.endswith(".dat"):
            self.filename += ".dat"

    def write(self, estimator):
        """TODO"""
        # save to single page to a file without number (i.e. output.dat)
        if len(estimator.pages) == 1:
            self.write_single_page(estimator.pages[0], self.filename)
        else:
            # split output path into directory, basename and extension
            dir_path = os.path.dirname(self.filename)
            if not os.path.exists(dir_path):
                logger.info("Creating {}".format(dir_path))
                os.makedirs(dir_path)
            file_base_part, file_ext = os.path.splitext(os.path.basename(self.filename))

            # loop over all pages and save an image for each of them
            for i, page in enumerate(estimator.pages):

                # calculate output filename. it will include page number padded with zeros.
                # for 10-99 pages the filename would look like: output_p01.png, ... output_p99.png
                # for 100-999 pages the filename would look like: output_p001.png, ... output_p999.png
                zero_padded_page_no = str(i + 1).zfill(len(str(len(estimator.pages))))
                output_filename = "{}_p{}{}".format(file_base_part, zero_padded_page_no, file_ext)
                output_path = os.path.join(dir_path, output_filename)

                # save the output file
                logger.info("Writing {}".format(output_path))
                self.write_single_page(page, output_path)

        return 0

    def write_single_page(self, page, filename):
        """TODO"""
        logger.info("Writing: " + filename)

        # special case for 0-dim data
        if page.dimension == 0:
            # save two numbers to the file
            if not np.all(np.isnan(page.error_raw)) and np.any(page.error_raw):
                np.savetxt(self.filename, [[page.data_raw, page.error_raw]], fmt="%g %g", delimiter=' ')
            else:  # save one number to the file
                np.savetxt(self.filename, [page.data_raw], fmt="%g", delimiter=' ')
        else:
            axis_numbers = list(range(page.dimension))

            # each axis may have different number of points, this is what we store here:
            axis_data_columns_1d = [page.plot_axis(i).data for i in axis_numbers]

            # now we calculate running index for each axis
            axis_data_columns_long = [np.meshgrid(*axis_data_columns_1d, indexing='ij')[i].ravel()
                                      for i in axis_numbers]

            fmt = "%g" + " %g" * page.dimension
            data_to_save = axis_data_columns_long + [page.data_raw]

            # if error information is present save it as additional column
            if not np.all(np.isnan(page.error_raw)) and np.any(page.error_raw):
                fmt += " %g"
                data_to_save += [page.error_raw]

            # transpose from rows to columns
            data_columns = np.transpose(data_to_save)

            # save space-delimited text file
            np.savetxt(filename, data_columns, fmt=fmt, delimiter=' ')
        return 0


class ImageWriter:
    """Writer responsible for creating PNG images using matplotlib library"""

    def __init__(self, filename, options):
        logger.info("{:s} options:  {:s}".format(repr(self.__class__), repr(options)))
        self.plot_filename = filename
        if not self.plot_filename.endswith(".png"):
            self.plot_filename += ".png"
        self.colormap = options.colormap
        self.axis_with_logscale = {PlotAxis[name] for name in options.log}

    default_colormap = 'gnuplot2'

    @staticmethod
    def _make_label(unit, name):
        """Make label for plot axis"""
        return name + " " + "[" + unit + "]"

    def get_page_figure(self, page):
        """Calculate matplotlib figure object for a single page in estimator"""
        try:
            import matplotlib
            matplotlib.use('Agg')
            import matplotlib.pyplot as plt
            from matplotlib import colors
            # set matplotlib logging level to ERROR, in order not to pollute our log space
            logging.getLogger('matplotlib').setLevel(logging.ERROR)
        except ImportError:
            logger.error("Matplotlib not installed, output won't be generated")
            return None

        # skip plotting 1-D and 3-D and higher dimensional data
        if page.dimension not in (1, 2):
            return None

        data_raw = page.data_raw
        error_raw = page.error_raw

        plot_x_axis = page.plot_axis(0)

        fig, ax = plt.subplots()
        ax.set_xlabel(self._make_label(plot_x_axis.unit, plot_x_axis.name))

        # we use symmetrical logarithmic scale as horizontal (X) axis for 1D and 2D plots
        # can have negative values as well (i.e. span from -4. to 4.)
        if PlotAxis.x in self.axis_with_logscale:
            ax.set_xscale('symlog')

        # 1-D plotting
        if page.dimension == 1:

            # scored values cannot be negative, hence we use purely logarithmic scale for vertical axis
            if PlotAxis.y in self.axis_with_logscale:
                ax.set_yscale('log')

            # add optional error area
            if np.any(page.error):
                ax.fill_between(plot_x_axis.data,
                                (data_raw - error_raw).clip(0.0),
                                (data_raw + error_raw).clip(0.0, 1.05 * data_raw.max()),
                                alpha=0.2, edgecolor='#CC4F1B', facecolor='#FF9848', antialiased=True)
            ax.set_ylabel(self._make_label(page.unit, page.name))
            ax.grid(True, alpha=0.3)
            ax.plot(plot_x_axis.data, data_raw)
        elif page.dimension == 2:
            plot_y_axis = page.plot_axis(1)

            x_axis_label = self._make_label(plot_x_axis.unit, plot_x_axis.name)
            y_axis_label = self._make_label(plot_y_axis.unit, plot_y_axis.name)
            z_axis_label = self._make_label(page.unit, page.name)

            # we use symmetrical logarithmic scale as vertical (Y) axis for 2D plots
            # can have negative values as well (i.e. span from -4. to 4.)
            if PlotAxis.y in self.axis_with_logscale:
                ax.set_yscale('symlog')

            # configure logscale on Z axis
            if PlotAxis.z in self.axis_with_logscale:
                norm = colors.LogNorm(vmin=data_raw[data_raw > 0].min(), vmax=data_raw.max())
            else:
                norm = colors.Normalize(vmin=data_raw.min(), vmax=data_raw.max())

            xspan = [plot_x_axis.min_val, plot_x_axis.max_val]
            yspan = [plot_y_axis.min_val, plot_y_axis.max_val]
            zdata = data_raw.reshape((plot_y_axis.n, plot_x_axis.n))

            plt.xlabel(x_axis_label)
            plt.ylabel(y_axis_label)

            plt.grid(True, alpha=0.3)

            im = ax.pcolorfast(xspan, yspan, zdata, cmap=self.colormap, norm=norm)
            cbar = plt.colorbar(im)
            if PlotAxis.z in self.axis_with_logscale:
                import matplotlib.ticker as ticker
                cbar.set_ticks(ticker.LogLocator(subs='all', numticks=15))
            cbar.set_label(z_axis_label, rotation=270, verticalalignment='bottom')

        return fig

    def write(self, estimator):
        """Go through all pages in estimator and save corresponding figure to an output file"""
        # save single page to a file without number (i.e. output.png)
        if len(estimator.pages) == 1:
            fig = self.get_page_figure(estimator.pages[0])
            if fig:
                logger.info("Writing {}".format(self.plot_filename))
                fig.savefig(self.plot_filename)
        else:
            # split output path into directory, basename and extension
            dir_path = os.path.dirname(self.plot_filename)
            if not os.path.exists(dir_path):
                logger.info("Creating {}".format(dir_path))
                os.makedirs(dir_path)
            file_base_part, file_ext = os.path.splitext(os.path.basename(self.plot_filename))

            # loop over all pages and save an image for each of them
            for i, page in enumerate(estimator.pages):

                # calculate output filename. it will include page number padded with zeros.
                # for 10-99 pages the filename would look like: output_p01.png, ... output_p99.png
                # for 100-999 pages the filename would look like: output_p001.png, ... output_p999.png
                zero_padded_page_no = str(i + 1).zfill(len(str(len(estimator.pages))))
                output_filename = "{}_p{}{}".format(file_base_part, zero_padded_page_no, file_ext)
                output_path = os.path.join(dir_path, output_filename)

                # save the output file
                fig = self.get_page_figure(page)
                if fig:
                    logger.info("Writing {}".format(output_path))
                    fig.savefig(output_path)

        return 0



import importlib
import importlib.util
import os


__version__ = 'unknown'

# versioning of the package is based on __version__ variable, declared here in __init__.py file
# __version__ can be either calculated based on pymchelper/VERSION file, storing version number
# or calculated directly from git repository (in case we work with cloned repo)

# we start by finding absolute path to the directory where pymchelper is installed
# we cannot rely on relative paths as binary files which need version number may be located in different
# directory than the pymchelper module

# get location of __init__.py file (the one you see now) in the filesystem
module_name = 'pymchelper'
spec = importlib.util.find_spec(module_name)
init_location = spec.origin

# VERSION should sit next to __init__.py in the directory structure
version_file = os.path.join(os.path.dirname(init_location), 'VERSION')

# let us try reading the VERSION file,
try:
    # it may be that executable (zipfile with pyz extension) is produced which contains all pymchelper
    # modules files packed together with __main__ script. In that case we cannot simply open file,
    # but we need to reach into the contents of zip file:
    possible_zipfile_path = os.path.normpath(os.path.join(init_location, '..', '..'))
    import zipfile
    if zipfile.is_zipfile(possible_zipfile_path):  # reach into zip file
        with zipfile.ZipFile(possible_zipfile_path, 'r') as zipfile:  # read file from inside ZIP archive
            # read first line of the file (removing newline character)
            vf = os.path.join(module_name, 'VERSION')
            with zipfile.open(vf, 'r') as f:
                __version__ = f.readline().strip().decode('ascii')
    # it looks that pymchelper was installed via pip and unpacked into some site-directory folder in the file system
    # in such case we simply read the VERSION file
    else:
        # read first line of the file (removing newline character)
        with open(version_file, 'r') as f:
            __version__ = f.readline().strip()

# in case above methods do not work, let us assume we are working with GIT repository
except FileNotFoundError:
    # backup solution - read the version from git
    from pymchelper.version import git_version
    __version__ = git_version()
